/*	$Id: pomExecTaskCodels.c 2010/04/19 17:25:12 mallet $ */

/* 
 * Copyright (c) 2000-2008,2010 LAAS/CNRS                  --  Thu Jun  8 2000
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <portLib.h>
#include <h2timeLib.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <math.h>
#include "h2mathLib.h"

#include "zlib.h"

#include "insitu/lib.h"
#include "libpom/lib.h"

/* module pom */
#include "pomCodels.h"
#include "server/pomHeader.h"
#include "server/pomEndian.h"

#include "libeuler/pomEuler.h"


/* --- Local data ---------------------------------------------------- */

/** Default fusion method for fused motion estimators */
#define POM_DEFAULT_FUSION_METHOD POM_MAX_CONF

/* Structure position de la libpom, uniquement pour calculs */
LibPomPos libpompos;
static int addMEFlag = 0;

/* Geometric model */
static InsituModelId	insituId;
static InsituFrameSet   frameSet;

/* structure contenant les id des PostersFrames. A allouer lors du chargement du modele. nbFrames contient le nombre de frames du modele, cad le nombre de posters pour le modele */
POM_FRAME_POSTER_ASSOC *frameAssocTab = NULL;

/* Me's and posterId's list */
POM_ME_POSTER_ASSOC *meAssocList = NULL;

/* Se's and posterId's list */
POM_SE_POSTER_ASSOC *seAssocList = NULL;


/* Error handling in the module */
#define pom_error(string, bilan, status) {		\
   fprintf(stderr,				\
	     "%s: S_pom_" #status "\n"		\
	     "%s: file %s, line %d\n",			\
	     string, string, __FILE__, __LINE__);	\
   errno = *(bilan) = S_pom_ ## status;		\
}

#define pom_error_param(string, param, bilan, status) {	\
   fprintf(stderr,				\
	     "%s: S_pom_" #status " %s\n"		\
	     "%s: file %s, line %d\n",			\
	     string, param, 				\
             string, __FILE__, __LINE__);			\
   errno = *(bilan) = S_pom_ ## status;		\
}


/* Local functions --------------------------------------------------- */

static STATUS	updatePoster(int dofusion, int *bilan);
static STATUS	pomPOM_SE_POSTERPosterReadOffset(POSTER_ID posterId,
			POM_SE_POSTER *x, size_t offset);


/*
 * pomExecTaskInit  -----------------------------------------------------
 *
 * Codel d'initialisation (SDI/f, ...)
 */

STATUS
pomExecTaskInit(int *bilan)
{
   SDI_F->mode = POM_STOP;
   SDI_F->fusionMethod = POM_DEFAULT_FUSION_METHOD;
   pomSetFusionMethod(POM_DEFAULT_FUSION_METHOD);

   /* Default values for CreateData */
   sprintf(SDI_F->medata.name,"%s",POMDEF_ME_NAME);
   SDI_F->medata.fuseFlag = POMDEF_ME_FUSE;

   sprintf(SDI_F->sedata.namese,"%s",POMDEF_SE_NAME);
   sprintf(SDI_F->sedata.namefrom,"%s",POMDEF_SE_FROM);
   sprintf(SDI_F->sedata.nameto,"%s",POMDEF_SE_TO);

   /*sprintf(SDI_F->modelname.name,"%s",POMDEF_MODEL_FILE);

   printf("\nModel name %s .... OK\n",SDI_F->modelname.name);
   */

   insituId = NULL;

   libpompos.date = 0;
   t3dInit(&libpompos.mainToBase, T3D_U_MATRIX, T3D_ALLOW_CONVERSION);
   t3dInit(&libpompos.mainToOrigin, T3D_U_MATRIX, T3D_ALLOW_CONVERSION);
   t3dInit(&libpompos.VLocal, T3D_U_MATRIX, T3D_ALLOW_CONVERSION);

   SDI_F->pomPos.date = 0;
   SDI_F->pomPos.pomTickDate = 0;
   SDI_F->pomPos.mainToBase=pomEulerVIdentity;
   SDI_F->pomPos.mainToOrigin=pomEulerVIdentity;
   SDI_F->pomPos.VLocal= pomEulerVIdentity;
   return OK;
}


/*
 * pomExecTaskEnd  ------------------------------------------------------
 *
 * Codel de terminaison
 */

STATUS
pomExecTaskEnd(void)
{
  int i;

  if (insituId != NULL) {
    /* Suppression des posters FramePos */
    for(i=0;i<frameSet.n;i++)
      posterDelete((frameAssocTab+i)->posterId);
    free(frameAssocTab);
    frameAssocTab = NULL;
    /* fermeture d'Insitu */
    insituClose(insituId);
    insituId = NULL;
  }

  return OK;
}


/*
 * pomExecTaskPerm  -----------------------------------------------------
 *
 * codel de l'activit permanente
 */

STATUS
pomExecTaskPerm(int *bilan)
{
   int dofusion;
   H2TIME date;
   STATUS status;
   LibPomMEPos libMEPos;
   POM_ME_POS pomMEPos;
   POM_ME_POSTER_ASSOC *me;
   POM_SE_POSTER_ASSOC *se;
   POM_SE_POSTER seposter;
   T3D tmp;

   if (SDI_F->mode != POM_START) return OK;
   
   /* insitu */
   if (insituId != NULL) {
     for(se = seAssocList; se != NULL; se = se->next) {
       status = pomPOM_SE_POSTERPosterReadOffset(se->posterId, &seposter, se->offset);
       if (status != OK) {
	 fprintf(stderr, "pom:%s:", se->name);
	 perror(NULL);
       } else {
	 pomEulerVToT3D(&seposter.seConfig, &tmp);
	 tmp.flags = T3D_ALLOW_CONVERSION;
	 if (insituSet(se->linkId, &tmp) != 0) {
	   fprintf(stderr, "pom:%s2%s:", se->from, se->to);
	   perror(NULL);
	 }
       }
     }
   } else {
     return OK;
   }

   /* pom */
   dofusion = 0;
   for(me = meAssocList; me != NULL; me = me->next) {
      status = posterIoctl(me->posterId, FIO_GETDATE, (char *)&date);
      if (status == OK && date.ntick != me->last.ntick) {
	 me->last = date;
	 if (pomPOM_ME_POSPosterRead(me->posterId, &pomMEPos) == ERROR) {
	    fprintf(stderr, "pom:%s:", me->name);
	    perror(NULL);
	 } else {
	    dofusion = 1;
	    if ( pomMEPos.kind == POM_ABSOLUTE_ME ) {
	      libMEPos.class = pomMEPos.kind;
	      libMEPos.posType = POM_ABSOLUTE_ME;
	      libMEPos.confidence = pomMEPos.confidence;
	      libMEPos.pos.abs.date = pomMEPos.date1;
	      pomEulerVToT3D(&pomMEPos.main, &tmp);
	      t3dSetOk(&libMEPos.pos.abs.mainToOrigin);
	      t3dSetOk(&tmp);
	      t3dCopyIn(&libMEPos.pos.abs.mainToOrigin,&tmp);
	      libMEPos.velocity.date = pomMEPos.date1;
	      pomEulerVToT3D(&pomMEPos.VLocal, &tmp);
	      t3dSetOk(&tmp);
	      t3dSetOk(&libMEPos.velocity.VLocal);
	      t3dCopyIn(&libMEPos.velocity.VLocal,&tmp);
	    } else {
	      libMEPos.class = pomMEPos.kind;
	      libMEPos.posType = POM_ABSOLUTE_ME;
	      libMEPos.confidence = pomMEPos.confidence;
	      libMEPos.pos.delta.date1 = pomMEPos.date1;
	      libMEPos.pos.delta.date2 = pomMEPos.date2;
	      pomEulerVToT3D(&pomMEPos.main, &tmp);
	      t3dSetOk(&libMEPos.pos.delta.main);
	      t3dSetOk(&tmp);
	      t3dCopyIn(&libMEPos.pos.delta.main,&tmp);
	      libMEPos.velocity.date = pomMEPos.date1;
	      pomEulerVToT3D(&pomMEPos.VLocal, &tmp);
	      t3dSetOk(&tmp);
	      t3dSetOk(&libMEPos.velocity.VLocal);
	      t3dCopyIn(&libMEPos.velocity.VLocal,&tmp);
	    }

	    if (pomMEUpdate(me->me, &libMEPos)) {
	      me->fuseOk = 0;
	      fprintf(stderr, "pomMEUpdate:%s:", me->name);
	      perror(NULL);
	    } else
	      me->fuseOk = 1;
	 }
      }
   }
   SDI_F->pomDate=pomGetTick();
   SDI_F->pomPos.pomTickDate = SDI_F->pomDate;

   /* update poster */
   if (updatePoster(dofusion, bilan) != OK) {
      perror("updatePoster");
      *bilan = OK;
   }
   return OK;
}

/*
 * Codel de creation des posters FramePos
 * ATTENTION: cette fonction ne renvoie rien et c'est normal. Elle est appelee automotiquement par genom,
 * hors on ne connait pas la geometrie du robot a l'avance. On utilisera alors dans les codels la fonction * suivante.
 */
STATUS pomCreatePosterFramePos(char *name, int size, POSTER_ID *pPosterId) {
  return OK;
}

STATUS pomRealCreateFramePoster(char *name, POSTER_ID *pPosterId,
				POM_SENSOR_POS *defaultFramePos)
{
  if (posterCreate(name, sizeof(POM_SENSOR_POS), pPosterId) != OK) {
    printf ("pomRealCreateFramePoster: cannot create poster frame %s %s\n", name, h2getMsgErrno(errnoGet()));
    return(ERROR);
  }

  if (pomFramePospom_sensor_posPosterWrite(*pPosterId, defaultFramePos) != OK) {
    printf ("pomRealCreateFramePoster: cannot write poster frame %s\n", name);
    return(ERROR);
  }

  return OK;
}



/*
 * SetModel -------------------------------------------------------------
 *
 */

/* pomSetModelExec  -  codel EXEC de SetModel
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
pomSetModelExec(POM_STRING *model, int *bilan)
{
  int i;
  InsituFrame *e;
  POM_SENSOR_POS initFramePosterVar;
  POSTER_ID posterId;
  char nametmp[ISFRM_MAX_LEN];
  size_t frameSetSize;

  /* Initialization of init_frame_poster */
  initFramePosterVar.date = 0;
  initFramePosterVar.pad = 0;
  initFramePosterVar.sensorToMain = pomEulerVIdentity;
  initFramePosterVar.mainToBase = pomEulerVIdentity;
  initFramePosterVar.VLocal = pomEulerVIdentity;

   /* If another model is loaded, does not make anything */
  if (insituId != NULL) {
    printf("Model already loaded, I try to reload it\n");
    SDI_F->mode = POM_STOP;
    insituClose(insituId);
    insituId = insituOpen(model->name, ISMODE_EDIT);
    if (insituId == NULL) pom_error("pomSetModelExec", bilan, NOT_ENOUGH_MEMORY );
    return ETHER;
   }

   /* Open the model */
   insituId = insituOpen(model->name, ISMODE_EDIT);
   if (insituId == NULL) {
     fprintf(stderr,"Can't open %s \n",model->name);
      SDI_F->mode = POM_STOP;
      switch(errno) {
	 case ENOMEM:
	    pom_error("pomSetModelExec", bilan, NOT_ENOUGH_MEMORY);
	    break;
	 default:
	    pom_error("pomSetModelExec", bilan, SYNTAX_ERROR );
	    break;
      }
      return ETHER;
   }

   /* for the first time, get robot internal configuration */
   /* get frameSetSize */
   frameSetSize = insituGetExportSize(insituId);
   /* allocate memory */
   frameSet.frame = (InsituFrame *) malloc (frameSetSize);
   if  (frameSet.frame == NULL) {
     pom_error("pomSetModelExec", bilan, NOT_ENOUGH_MEMORY );
     return ETHER;
   }
   if (insituExport(insituId, &frameSet) < 0) {
     fprintf(stderr, "error while exporting frame '%s'\n",
	     frameSet.frame[frameSet.n-1].name);
     pom_error("insituExport", bilan, CANNOT_GET_FRAMES);
     return ETHER;
   }


   /* Initialisation de frameAssocList et creation des posters */
   e = &frameSet.frame[0];
   if ((frameAssocTab = malloc(frameSet.n * sizeof(POM_FRAME_POSTER_ASSOC))) == NULL){
       pom_error("pomSetModelExec", bilan, NOT_ENOUGH_MEMORY );
       return ETHER;
     }

   for(i=0;i<frameSet.n;i++,e++){
     /* ecriture du nom du poster */
     sprintf(nametmp, POM_FRAME_POSTER_NAME, e->name);
     strncpy((frameAssocTab+i)->name, nametmp, ISFRM_MAX_LEN);
     /* recuperation des valeurs de la frame */
     initFramePosterVar.date = 0;
     t3dConvertTo(T3D_EULER,&e->frameToMain);

     pomEulerVFromT3D(&e->frameToMain, &initFramePosterVar.sensorToMain);
     initFramePosterVar.mainToBase = SDI_F->pomPos.mainToBase;
     initFramePosterVar.mainToOrigin = SDI_F->pomPos.mainToOrigin;
     initFramePosterVar.VLocal = SDI_F->pomPos.VLocal;

     /* creation du poster */
     if (pomRealCreateFramePoster(nametmp, &posterId, &initFramePosterVar) == ERROR) {
       pom_error_param("pomSetModelExec", nametmp, bilan, CANNOT_CREATE_POSTER );
       return ETHER;
     }
     (frameAssocTab+i)->posterId = posterId;
     printf("create poster %s\n",(frameAssocTab+i)->name);
   }

   return ETHER;
}

/*------------------------------------------------------------------------
 * MonitorPosKO
 *
 * Description:
 *
 * Bilans:      OK
 *              S_pom_BAD_PARAMS
 *              S_pom_POM_X_OUT
 *              S_pom_POM_Y_OUT
 *              S_pom_POM_Z_OUT
 *              S_pom_POM_THETA_OUT
 *              S_pom_POM_PHI_OUT
 *              S_pom_POM_PSI_OUT
 */

/* pomMonitorPosKOExec  -  codel EXEC de MonitorPosKO
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT pomMonitorPosKOExec (POM_UNC *pomUnc, int *bilan)
{
  double errTheta;
  errTheta = POM_3SIGMA(SDI_F->pomPos.mainToOrigin.var.voo);
  if (errTheta < 3 && errTheta > pomUnc->utheta) {
    *bilan = S_pom_POM_THETA_OUT;
    return ETHER;
  }
  /* if (POM_3SIGMA(SDI_F->pomPos.mainToOrigin.var.vpp) > pomUnc->uphi) { */
/*     *bilan = S_pom_POM_PHI_OUT; */
/*     return ETHER; */
/*   } */
/*   if (POM_3SIGMA(SDI_F->pomPos.mainToOrigin.var.vrr) > pomUnc->upsi) { */
/*     *bilan = S_pom_POM_PSI_OUT; */
/*     return ETHER; */
/*   } */
  if (POM_3SIGMA(SDI_F->pomPos.mainToOrigin.var.vxx) > pomUnc->ux) {
    *bilan = S_pom_POM_X_OUT;
    return ETHER;
  }
  if (POM_3SIGMA(SDI_F->pomPos.mainToOrigin.var.vyy) > pomUnc->uy) {
    *bilan = S_pom_POM_Y_OUT;
    return ETHER;
  }
  /* if (POM_3SIGMA(SDI_F->pomPos.mainToOrigin.var.vzz) > pomUnc->uz) { */
/*     *bilan = S_pom_POM_Z_OUT; */
/*     return ETHER; */
/*   } */

  return EXEC;
}

/*------------------------------------------------------------------------
 * MonitorPosOK
 *
 * Description:
 *
 * Bilans:      OK
 *              S_pom_BAD_PARAMS
 */

/* pomMonitorPosOKExec  -  codel EXEC de MonitorPosOK
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT pomMonitorPosOKExec (POM_UNC *pomUnc, int *bilan)
{
  if (POM_3SIGMA(SDI_F->pomPos.mainToOrigin.var.voo) < pomUnc->utheta &&
      /* POM_3SIGMA(SDI_F->pomPos.mainToOrigin.var.vpp) < pomUnc->uphi && */
/*       POM_3SIGMA(SDI_F->pomPos.mainToOrigin.var.vrr) < pomUnc->upsi && */
      POM_3SIGMA(SDI_F->pomPos.mainToOrigin.var.vxx) < pomUnc->ux &&
      POM_3SIGMA(SDI_F->pomPos.mainToOrigin.var.vyy) < pomUnc->uy /* && */
/*       POM_3SIGMA(SDI_F->pomPos.mainToOrigin.var.vzz) < pomUnc->uz */ )
    return ETHER;
  return EXEC;
}

/*------------------------------------------------------------------------
 * MonitorMEReject
 *
 * Description:
 *
 * Bilans:      OK
 *              S_pom_ME_REJECTED
 */

/* pomMonitorMERejectExec  -  codel EXEC de MonitorMEReject
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT pomMonitorMERejectExec (int *bilan)
{
  POM_ME_POSTER_ASSOC *me;

  for(me = meAssocList; me != NULL; me = me->next) {
    if (!me->fuseOk) {
      *bilan = S_pom_ME_REJECTED;
      me->fuseOk = 1;
      return ETHER;
    }
  }
  return EXEC;
}


/*
 * AddME ----------------------------------------------------------------
 *
 */

/* pomAddMEExec  -  codel EXEC de AddME
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
pomAddMEExec(POM_ME_CREATE_DATA *data, int *bilan)
{
   POM_ME_POSTER_ASSOC *meAssoc;
   int me;

   me = pomMECreate(data->name, (data->fuseFlag == POM_FUSE)?1:0);
   if (me < 0) {
      switch(errno) {
	 case EINVAL:
	    printf("ME %s already exists\n",data->name);
	    pom_error("pomAddMEExec", bilan, ME_ALREADY_EXISTS);
	    break;
	 case ENOMEM:
	 default:

	    pom_error("pomAddMEExec", bilan, NOT_ENOUGH_MEMORY);
	    break;
      }
      return ETHER;
   }

   for(meAssoc = meAssocList; meAssoc != NULL; meAssoc = meAssoc->next)
      if (!strncmp(data->name, meAssoc->name, POM_ME_NAME_MAXLEN))
	 break;

   if (meAssoc == NULL) {
      meAssoc = malloc(sizeof(POM_ME_POSTER_ASSOC));
      if (meAssoc == NULL) {
	 pom_error("pomAddMEExec", bilan, NOT_ENOUGH_MEMORY);
	 return ETHER;
      }

      meAssoc->next = meAssocList;
      meAssocList = meAssoc;
   }
   meAssoc->last.ntick = -1;
   strncpy(meAssoc->name, data->name, POM_ME_NAME_MAXLEN);
   meAssoc->me = me;
   meAssoc->fuse = (data->fuseFlag == POM_FUSE)?1:0;
   meAssoc->fuseOk = 1;

   sprintf(SDI_F->tmp, POM_ME_POSTER_NAME, data->name);
   if (posterFind(SDI_F->tmp, &meAssoc->posterId) != OK) {
     meAssocList = meAssoc->next;
     free(meAssoc);
     pom_error_param("pomAddMEExec", SDI_F->tmp, bilan, CANNOT_FIND_POSTER);
     return ETHER;
   }

   addMEFlag = 1;
   return ETHER;
}


/*
 * RemoveME -------------------------------------------------------------
 *
 */

/* pomRemoveMEExec  -  codel EXEC de RemoveME
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
pomRemoveMEExec(POM_STRING *name, int *bilan)
{
   POM_ME_POSTER_ASSOC *me, *tmp;

   me = NULL;

   if (meAssocList == NULL) {
      pom_error_param("pomAddMEExec", name->name, bilan, NO_SUCH_ME);
      return ETHER;
   }

   if (!strncmp(name->name, meAssocList->name, POM_ME_NAME_MAXLEN)) {
      me = meAssocList;
      meAssocList = meAssocList->next;
   } else {
      for(tmp = meAssocList; tmp->next != NULL; tmp = tmp->next)
	 if (!strncmp(name->name, tmp->next->name, POM_ME_NAME_MAXLEN)) {
	    me = tmp->next;
	    break;
	 }
      if (me != NULL) tmp->next = me->next;
   }

   if (me == NULL) {
      pom_error_param("pomAddMEExec", name->name, bilan, NO_SUCH_ME);
      return ETHER;
   }

   pomMERemove(me->me);
   free(me);
   return ETHER;
}

/*------------------------------------------------------------------------
 * Reset
 *
 * Description: go in POM_STOP mode and reset all MEs.
 *
 * Bilans:      OK
 */

/* pomResetExec  -  codel EXEC de Reset
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT pomResetExec (int *bilan)
{
  POM_ME_POSTER_ASSOC *me, *tmp;

  SDI_F->mode = POM_STOP;
  for(me = meAssocList; me != NULL; me = tmp) {
    tmp = me->next;
    pomMERemove(me->me);
    free(me);
  }
  pomReset();
  libpompos.date = 0;
   t3dInit(&libpompos.mainToBase, T3D_U_MATRIX, T3D_ALLOW_CONVERSION);
   t3dInit(&libpompos.mainToOrigin, T3D_U_MATRIX, T3D_ALLOW_CONVERSION);
   t3dInit(&libpompos.VLocal, T3D_U_MATRIX, T3D_ALLOW_CONVERSION);
  meAssocList = NULL;
  addMEFlag = 0;
  return ETHER;
}


/*------------------------------------------------------------------------
 * AddSE
 *
 * Description:
 *
 * Bilans:      OK
 *              S_pom_NO_INSITU
 *              S_pom_CANNOT_FIND_LINK
 *              S_pom_ME_ALREADY_EXISTS
 *              S_pom_CANNOT_FIND_POSTER
 *              S_pom_NOT_ENOUGH_MEMORY
 */

/* pomAddSEExec  -  codel EXEC de AddSE
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT pomAddSEExec (POM_SE_CREATE_DATA *sedata, int *bilan)
{
  POM_SE_POSTER seData;
  POM_SE_POSTER_ASSOC *seAssoc;
  unsigned int i, size;

  if (insituId == NULL) {
      pom_error("pomAddSEExec", bilan, NO_INSITU);
      return ETHER;
   }

  for(seAssoc = seAssocList; seAssoc != NULL; seAssoc = seAssoc->next)
    if (!strncmp(sedata->namese, seAssoc->name, POM_SE_NAME_MAXLEN))
      break;

  if (seAssoc == NULL) {
    seAssoc = malloc(sizeof(POM_SE_POSTER_ASSOC));
    if (seAssoc == NULL) {
      pom_error("pomAddSEExec", bilan, NOT_ENOUGH_MEMORY);
      return ETHER;
    }

    seAssoc->next = seAssocList;
    seAssocList = seAssoc;
  }

  strncpy(seAssoc->name, sedata->namese, POM_SE_NAME_MAXLEN);
  strncpy(seAssoc->from, sedata->namefrom, ISFRM_MAX_LEN);
  strncpy(seAssoc->to, sedata->nameto, ISFRM_MAX_LEN);

  /* look for poster, get size and number of entries */

  if (posterFind(sedata->nameposter, &seAssoc->posterId) != OK) {
    pom_error_param("pomAddSEExec", sedata->nameposter, 
		    bilan, CANNOT_FIND_POSTER);
    goto error;
  }

  if (posterIoctl(seAssoc->posterId, FIO_GETSIZE, &size) != OK) {
    pom_error_param("size of poster is unknown", sedata->nameposter, bilan, CANNOT_FIND_POSTER);
    goto error;
  }
  if (size % sizeof(POM_SE_POSTER)) {
    pom_error_param("size of poster is invalid", sedata->nameposter, bilan, CANNOT_FIND_POSTER);
    goto error;
  }

  /* scan poster entries to find SE name */
  seAssoc->offset = -1;
  for(i=0; i < size; i+= sizeof(POM_SE_POSTER)) {
    if (pomPOM_SE_POSTERPosterReadOffset(seAssoc->posterId, &seData, i) != OK) {
      pom_error_param("cannot read poster", sedata->nameposter, bilan, CANNOT_FIND_POSTER);
      goto error;
    }

    if (!strcmp(seAssoc->name, seData.seName)) {
      seAssoc->offset = i;
      break;
    }
  }
  if (seAssoc->offset == -1) {
    pom_error_param("cannot find SE name", seAssoc->name, bilan, CANNOT_FIND_POSTER);
    goto error;
  }

  /* get insitu id */
  if ((seAssoc->linkId = insituOpenLink(insituId, seAssoc->from, seAssoc->to)) == NULL) {
    seAssocList = seAssoc->next;
    free(seAssoc);
    pom_error_param("pomAddSEExec", seAssoc->from, bilan, CANNOT_FIND_LINK);
    return ETHER;
  }

  return ETHER;

error:
  seAssocList = seAssoc->next;
  free(seAssoc);
  return ETHER;
}

/*------------------------------------------------------------------------
 * RemoveSE
 *
 * Description:
 *
 * Bilans:      OK
 *              S_pom_NO_SUCH_SE
 */

/* pomRemoveSEExec  -  codel EXEC de RemoveSE
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT pomRemoveSEExec (POM_STRING *name, int *bilan)
{
  POM_SE_POSTER_ASSOC *se, *tmp;

   se = NULL;

   if (seAssocList == NULL) {
      pom_error("pomRemoveSEExec", bilan, NO_SUCH_ME);
      return ETHER;
   }

   if (!strncmp(name->name, seAssocList->name, POM_SE_NAME_MAXLEN)) {
      se = seAssocList;
      seAssocList = seAssocList->next;
   } else {
      for(tmp = seAssocList; tmp->next != NULL; tmp = tmp->next)
	 if (!strncmp(name->name, tmp->next->name, POM_SE_NAME_MAXLEN)) {
	    se = tmp->next;
	    break;
	 }
      if (se != NULL) tmp->next = se->next;
   }

   if (se == NULL) {
      pom_error("pomRemoveSEExec", bilan, NO_SUCH_ME);
      return ETHER;
   }

   free(se);
   return ETHER;
}

/*
 * SetRefME -------------------------------------------------------------
 *
 */

/* pomSetRefMEExec  -  codel EXEC de SetRefME
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
pomSetRefMEExec(POM_STRING *name, int *bilan)
{
   POM_ME_POSTER_ASSOC *me;

   for(me = meAssocList; me != NULL; me = me->next)
      if (!strncmp(name->name, me->name, POM_ME_NAME_MAXLEN))
	 break;

   if (me == NULL) {
      pom_error("pomSetRefME", bilan, NO_SUCH_ME);
      return ETHER;
   }

   if (pomSetRefME(me->me)) {
      pom_error("pomSetRefME", bilan, NO_SUCH_ME);
      return ETHER;
   }

   return ETHER;
}


/*
 * SetPos ---------------------------------------------------------------
 *
 */

/* pomSetPosExec  -  codel EXEC de SetPos
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
pomSetPosExec(POM_SET_POS *pos, int *bilan)
{
   int i;
   InsituFrame *e;
   T3D m, f;

   if (insituId == NULL) {
      pom_error("pomSetPosExec", bilan, NO_INSITU);
      return ETHER;
   }

   /* look for frame (in internal structs) */
   e = &frameSet.frame[0];
   for(i=0;i<frameSet.n;i++,e++)
      if (!strcmp(pos->frame, e->name)) break;
   if (i >= frameSet.n) {
      pom_error("pomSetPosExec", bilan, UNKNOWN_FRAME);
      return ETHER;
   }

   /* compute "main to origin" (the given pos is "frame to origin") */
   t3dCopyIn(&m, &e->frameToMain);
   t3dInvert(&m);

   t3dInit(&f, T3D_U_BRYAN, T3D_ALLOW_CONVERSION);
   f.bryan.yaw = pos->yaw;
   f.bryan.pitch = pos->pitch;
   f.bryan.roll = pos->roll;
   f.bryan.x = pos->x;
   f.bryan.y = pos->y;
   f.bryan.z = pos->z;

   t3dSetOk(&m);
   t3dConvertTo(T3D_U_BRYAN, &m);
   m.euler.flags |= T3D_ALLOW_CONVERSION;
   t3dComp(&m, &f);

   /* set position */
   if (pomSetZero(&m)) {
      pom_error("pomSetPosExec", bilan, CANNOT_SET_POS);
      return ETHER;
   }

   return ETHER;
}


/*
 * GetPos ---------------------------------------------------------------
 *
 */

/* pomGetPosExec  -  codel EXEC de GetPos
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
pomGetPosExec(POM_SET_POS *pos, int *bilan)
{
   T3D m;

   /* return "main to origin" */
   t3dSetOk(&m);
   t3dCopyIn(&m, &libpompos.mainToOrigin);
   m.euler.flags |= T3D_ALLOW_CONVERSION;
   t3dConvertTo(T3D_EULER, &m);

   strcpy(pos->frame, "mainToOrigin");
   /*frame is usually frameToMain, here it is not...*/
   pos->yaw = m.euler.euler[0];
   pos->pitch = m.euler.euler[1];
   pos->roll = m.euler.euler[2];
   pos->x = m.euler.euler[3];
   pos->y = m.euler.euler[4];
   pos->z = m.euler.euler[5];

   return ETHER;
}

/* pomGetFramePosExec  -  codel EXEC de GetFramePos
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
pomGetFramePosExec(POM_STRING *name, POM_SET_POS *pos, int *bilan)
{
   int i;
   InsituFrame *e;
   T3D m;

   if (insituId == NULL) {
      pom_error("pomGetPosExec", bilan, NO_INSITU);
      return ETHER;
   }

   /* look for frame (in internal structs) */
   e = &frameSet.frame[0];
   for(i=0;i<frameSet.n;i++,e++)
      if (!strcmp(name->name, e->name)) break;
   if (i >= frameSet.n) {
      pom_error("pomGetPosExec", bilan, UNKNOWN_FRAME);
      return ETHER;
   }

   /* return "frame to main" */
   t3dSetOk(&m);
   t3dCopyIn(&m, &e->frameToMain);
   m.euler.flags |= T3D_ALLOW_CONVERSION;
   t3dConvertTo(T3D_EULER, &m);

   strcpy(pos->frame, e->name);
   pos->yaw = m.euler.euler[0];
   pos->pitch = m.euler.euler[1];
   pos->roll = m.euler.euler[2];
   pos->x = m.euler.euler[3];
   pos->y = m.euler.euler[4];
   pos->z = m.euler.euler[5];

   return ETHER;
}


ACTIVITY_EVENT
pomSetAbsPosFromMEExec(POM_STRING *name,int *bilan)
{
  POM_ME_POSTER_ASSOC *me, *tmp;

  me = NULL;

  if (!strncmp(name->name, meAssocList->name, POM_ME_NAME_MAXLEN)) {
    me = meAssocList;
  } else {
    for(tmp = meAssocList; tmp->next != NULL; tmp = tmp->next)
      if (!strncmp(name->name, tmp->next->name, POM_ME_NAME_MAXLEN)) {
	me = tmp->next;
	break;
      }
  }

  if (me == NULL) {
    pom_error("pomSetAbsPosFromMEExec", bilan, NO_SUCH_ME);
    return ETHER;
  }

  if (pomClass(me->me)!=POM_ABSOLUTE_ME) {
    pom_error("pomSetAbsPosFromMEExec", bilan, NO_SUCH_ME);
    return ETHER;
  }
  fprintf(stderr,"Setting new offset from %s\n",name->name);
  if (pomSetPosFromME(me->me)) {
    pom_error("pomSetAbsPosFromMEExec", bilan, CANNOT_SET_POS);
    return ETHER;
  }

  return ETHER;
}


/* --- Log ----------------------------------------------------------- */

/** \brief Dump positions in a file
 *
 * EXEC codel of the Log request
 *
 * \return  EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
pomLogExec(POM_STRING *name, int *bilan)
{
   gzFile *f;

   /* empty name for disabling logging */
   if (!strcmp(name->name, "none") || name->name[0] == '\0') {
     f = pomLog(NULL);
     if (f != NULL) gzclose(f);
     printf("pom: finished logging\n");
     return ETHER;
   }

   /* To start a new log */
   f = gzopen(name->name, "w9");
   if (f == NULL) {
     pom_error("pomSetPosExec", bilan, CANNOT_LOG);
     return ETHER;
   }

   pomLog(f);
   printf("pom: logging to %s\n", name->name);
   return ETHER;
}

/** \brief Stop dumping positions in a file
 *
 * EXEC codel of the StopLog request
 *
 * \return  EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
pomStopLogExec(int *bilan)
{
  gzFile *f;
   
  f = pomLog(NULL);
  if (f != NULL) 
    gzclose(f);
  else
    pom_error("pomSetPosExec", bilan, CANNOT_STOP_LOG);
  
  return ETHER;
}

/*------------------------------------------------------------------------
 * DisplayFrame
 *
 * Description:
 *
 * Bilans:      OK
 *              S_pom_NO_FRAME_LIST
 */

/* pomDisplayFrameExec  -  codel EXEC de DisplayFrame
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT pomDisplayFrameExec (POM_STRING *name, POM_SENSOR_POS *framePos, int *bilan)
{
  int i=0;

  /* Display des FramePos */
    if (frameAssocTab == NULL){
      pom_error("pomDisplayFrameExec", bilan, CANNOT_GET_FRAMES );
      return ETHER;
    }

    while ( i < frameSet.n && strncmp(name->name,(frameAssocTab+i)->name,ISFRM_MAX_LEN))
      i++;

    if ((i == frameSet.n) || (posterRead((frameAssocTab+i)->posterId,0,(void *)framePos,
					 sizeof(POM_SENSOR_POS)) != sizeof(POM_SENSOR_POS))){
      pom_error("pomDisplayFrameExec", bilan, CANNOT_GET_FRAMES );
      return ETHER;
    }

  return ETHER;
}

/*------------------------------------------------------------------------
 * WriteSensorPosFile
 *
 * Description: 
 *
 * Reports:      OK
 *              S_pom_NO_FRAME_LIST
 *              S_pom_NO_SUCH_FRAME
 *              S_pom_CANNOT_OPEN_FILE
 */

/* pomWriteSensorPosFileMain  -  codel EXEC of WriteSensorPosFile
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
pomWriteSensorPosFileMain(POM_STRING_2 *posterName_fileName, int *report)
{
  int i=0;
  FILE *fileptr;
  POM_SENSOR_POS framePos;

  printf("Poster Name %s\n",posterName_fileName->name1);
  printf("File Name %s\n",posterName_fileName->name2);
  /* Display des FramePos */
  if (frameAssocTab == NULL){
    pom_error("pomWriteSensorPosFileMain", report, CANNOT_GET_FRAMES );
    return ETHER;
  }
  
  while ( i < frameSet.n && strncmp(posterName_fileName->name1,(frameAssocTab+i)->name,ISFRM_MAX_LEN))
    i++;
  
  if ((i == frameSet.n) || (posterRead((frameAssocTab+i)->posterId,0,(void *)&framePos,
				       sizeof(POM_SENSOR_POS)) != sizeof(POM_SENSOR_POS))){
    pom_error("pomWriteSensorPosFileMain", report, CANNOT_GET_FRAMES );
    return ETHER;
  }
 
  if ((fileptr = fopen(posterName_fileName->name2,"wb"))==NULL) {
    pom_error("pomWriteSensorPosFileMain", report, CANNOT_OPEN_FILE);
    return ETHER;
  }
  fprintf(fileptr,"date: %d\n",framePos.date);
  fprintf(fileptr,"sensorToMain: [6](%3.6f,%3.6f,%3.6f,%3.6f,%3.6f,%3.6f)\n",framePos.sensorToMain.euler.x,framePos.sensorToMain.euler.y, framePos.sensorToMain.euler.z, framePos.sensorToMain.euler.yaw, framePos.sensorToMain.euler.pitch, framePos.sensorToMain.euler.roll);
  fprintf(fileptr,"mainToBase: [6](%3.6f,%3.6f,%3.6f,%3.6f,%3.6f,%3.6f)\n",framePos.mainToBase.euler.x,framePos.mainToBase.euler.y, framePos.mainToBase.euler.z, framePos.mainToBase.euler.yaw, framePos.mainToBase.euler.pitch, framePos.mainToBase.euler.roll);
  fprintf(fileptr,"mainToOrigin: [6](%3.6f,%3.6f,%3.6f,%3.6f,%3.6f,%3.6f)\n",framePos.mainToOrigin.euler.x,framePos.mainToOrigin.euler.y, framePos.mainToOrigin.euler.z, framePos.mainToOrigin.euler.yaw, framePos.mainToOrigin.euler.pitch, framePos.mainToOrigin.euler.roll);

  fclose(fileptr);
  return ETHER;
}

static BOOL testJump(T3D *new, T3D *last,
		     double seuilCoord, double seuilAngle,
		     double coeffMultCoord, double coeffMultAngle) {

  double uth,uph,ups,ux,uy,uz;
  double eth,eph,eps,ex,ey,ez;

  uth = POM_3SIGMA(last->euler.v[0]) * coeffMultCoord + seuilCoord;
  uph = POM_3SIGMA(last->euler.v[2]) * coeffMultCoord + seuilCoord;
  ups = POM_3SIGMA(last->euler.v[5]) * coeffMultCoord + seuilCoord;
  ux = POM_3SIGMA(last->euler.v[9])  * coeffMultAngle + seuilAngle;
  uy = POM_3SIGMA(last->euler.v[14]) * coeffMultAngle + seuilAngle;
  uz = POM_3SIGMA(last->euler.v[20]) * coeffMultAngle + seuilAngle;

  eth = angleLimit(new->euler.euler[0]-last->euler.euler[0]);
  eph = angleLimit(new->euler.euler[1]-last->euler.euler[1]);
  eps = angleLimit(new->euler.euler[2]-last->euler.euler[2]);
  ex = new->euler.euler[3] - last->euler.euler[3];
  ey = new->euler.euler[4] - last->euler.euler[4];
  ez = new->euler.euler[5] - last->euler.euler[5];

  if (fabs(ex)  > (ux) ||
      fabs(ey)  > (uy) ||
      fabs(ez)  > (uz) ||
      fabs(eth) > (uth) ||
      fabs(eph) > (uph) ||
      fabs(eps) > (ups)) {

    printf ("Jump : dx %5.3f  (max %5.3f)\n"
	    "       dy %5.3f  (max %5.3f)\n"
	    "       dz %5.3f  (max %5.3f)\n"
	    "       dth %5.3f (max %5.3f)\n"
	    "       dph %5.3f (max %5.3f)\n"
	    "       dps %5.3f (max %5.3f)\n",
	    ex, ux, ey, uy, ez, uz,
	    eth, uth, eph, uph, eps, ups);
    /* XXXXXXX */
    return  1;
  }

  return 0;
}

/*
 * updatePoster ---------------------------------------------------------
 * On update egalement tous les FramePosters
 *
 */

static STATUS
updatePoster(int dofusion, int *bilan)
{
  int i;
  InsituFrame *e;
  POM_SENSOR_POS framePosterVar;
  static T3D lastMainToOrigin;

  /* compute best position */
  if (dofusion) {
	if (pomFuse(&libpompos))
		pom_error("pomFuse", bilan, CANNOT_FUSE_POSITIONS);
  }
  t3dConvertTo(T3D_EULER,&libpompos.mainToOrigin);

/*    fprintf(stderr,"updatePoster fusion:\n"); */
/*    dm_print(libpompos.mainToOrigin.euler.euler,1,6); */

  /* test not JUMP */
  if (!lastMainToOrigin.euler.err || addMEFlag) {
    addMEFlag = 0;
  }
  lastMainToOrigin.euler = libpompos.mainToOrigin.euler;

  /* get robot internal configuration */
  if ((*bilan==OK)&&(insituId == NULL ||
		     insituExport(insituId, &frameSet) < 0)) {
    pom_error("insituExport", bilan, CANNOT_GET_FRAMES);
    return ERROR;
  }

  /* increment date counter */
  if (pomTick() < 0) {
    pom_error("pomTick", bilan, CANNOT_FUSE_POSITIONS);
    return ERROR;
  }

  if (*bilan==OK) {

    /* Mise a jour de la SDI_F pour la position */
    SDI_F->pomPos.date = pomGetTick();
    t3dConvertTo(T3D_EULER, &libpompos.mainToBase);
    t3dConvertTo(T3D_EULER, &libpompos.mainToOrigin);
    t3dConvertTo(T3D_EULER, &libpompos.VLocal);

    pomEulerVFromT3D(&libpompos.mainToBase, &SDI_F->pomPos.mainToBase);
    pomEulerVFromT3D(&libpompos.mainToOrigin, &SDI_F->pomPos.mainToOrigin);
    pomEulerVFromT3D(&libpompos.VLocal, &SDI_F->pomPos.VLocal);

    /* Mise a jour des FramePos */
    e = &frameSet.frame[0];
    if (frameAssocTab == NULL){
      pom_error("pomSetModelExec", bilan, CANNOT_GET_FRAMES );
      return ERROR;
    }

    for(i=0;i<frameSet.n;i++,e++){
      /* recuperation des valeurs de la frame */
      framePosterVar.date = SDI_F->pomPos.date;
      t3dConvertTo(T3D_EULER,&e->frameToMain);
      pomEulerVFromT3D(&e->frameToMain, &framePosterVar.sensorToMain);
      framePosterVar.mainToBase = SDI_F->pomPos.mainToBase;
      framePosterVar.mainToOrigin = SDI_F->pomPos.mainToOrigin;
      framePosterVar.VLocal = SDI_F->pomPos.VLocal;
      /* ecriture dans le poster */
      if (pomFramePospom_sensor_posPosterWrite((frameAssocTab+i)->posterId, &framePosterVar) != OK) {
	printf ("pomUpdatePoster: cannot write poster frame %s\n",(frameAssocTab+i)->name);
	return ERROR;
      }
    }

  }
/*   printf("\n--\n"); */

  return OK;
}


/* --  pomPOM_SE_POSTERPosterReadOffset ------------------------------ */

/** \brief Read a POM_EULER_V poster.
 *
 * Read in the given \a posterId starting from \a offset
 *
 * \param[in] posterId	poster identifier
 * \param[out] x	buffer where data is read
 * \param[in] offset	offset inside \a posterId
 *
 * \return OK or ERROR
 */

static STATUS
pomPOM_SE_POSTERPosterReadOffset(POSTER_ID posterId, POM_SE_POSTER *x, size_t offset)
{
  H2_ENDIANNESS posterDataEndianness;
  if (posterEndianness(posterId, &posterDataEndianness) == ERROR) {
    h2perror("pomPOM_EULER_VPosterRead endianness");
    return ERROR;
  }
  if (posterRead(posterId, offset, (char *)x, sizeof(*x))
           != sizeof(*x)) {
    h2perror("pomPOM_SE_POSTERPosterReadOffset read");
    return ERROR;
  }
  if (posterDataEndianness != H2_LOCAL_ENDIANNESS)
    endianswap_struct_POM_SE_POSTER(x, 0, NULL);

  return OK;
}
